package ui

import "git.sr.ht/~rockorager/vaxis"

var KeyNames = map[string]rune{
	"Up":                   vaxis.KeyUp,
	"Right":                vaxis.KeyRight,
	"Down":                 vaxis.KeyDown,
	"Left":                 vaxis.KeyLeft,
	"Insert":               vaxis.KeyInsert,
	"Delete":               vaxis.KeyDelete,
	"BackSpace":            vaxis.KeyBackspace,
	"Page_Down":            vaxis.KeyPgDown,
	"Page_Up":              vaxis.KeyPgUp,
	"Home":                 vaxis.KeyHome,
	"End":                  vaxis.KeyEnd,
	"F0":                   vaxis.KeyF00,
	"F1":                   vaxis.KeyF01,
	"F2":                   vaxis.KeyF02,
	"F3":                   vaxis.KeyF03,
	"F4":                   vaxis.KeyF04,
	"F5":                   vaxis.KeyF05,
	"F6":                   vaxis.KeyF06,
	"F7":                   vaxis.KeyF07,
	"F8":                   vaxis.KeyF08,
	"F9":                   vaxis.KeyF09,
	"F10":                  vaxis.KeyF10,
	"F11":                  vaxis.KeyF11,
	"F12":                  vaxis.KeyF12,
	"F13":                  vaxis.KeyF13,
	"F14":                  vaxis.KeyF14,
	"F15":                  vaxis.KeyF15,
	"F16":                  vaxis.KeyF16,
	"F17":                  vaxis.KeyF17,
	"F18":                  vaxis.KeyF18,
	"F19":                  vaxis.KeyF19,
	"F20":                  vaxis.KeyF20,
	"F21":                  vaxis.KeyF21,
	"F22":                  vaxis.KeyF22,
	"F23":                  vaxis.KeyF23,
	"F24":                  vaxis.KeyF24,
	"F25":                  vaxis.KeyF25,
	"F26":                  vaxis.KeyF26,
	"F27":                  vaxis.KeyF27,
	"F28":                  vaxis.KeyF28,
	"F29":                  vaxis.KeyF29,
	"F30":                  vaxis.KeyF30,
	"F31":                  vaxis.KeyF31,
	"F32":                  vaxis.KeyF32,
	"F33":                  vaxis.KeyF33,
	"F34":                  vaxis.KeyF34,
	"F35":                  vaxis.KeyF35,
	"Clear":                vaxis.KeyClear,
	"Begin":                vaxis.KeyBegin,
	"Cancel":               vaxis.KeyCancel,
	"Print":                vaxis.KeyPrint,
	"Caps_Lock":            vaxis.KeyCapsLock,
	"Scroll_Lock":          vaxis.KeyScrollLock,
	"Num_Lock":             vaxis.KeyNumlock,
	"Pause":                vaxis.KeyPause,
	"Menu":                 vaxis.KeyMenu,
	"XF86AudioPlay":        vaxis.KeyMediaPlay,
	"XF86AudioPause":       vaxis.KeyMediaPause,
	"XF86AudioStop":        vaxis.KeyMediaStop,
	"XF86AudioForward":     vaxis.KeyMediaFF,
	"XF86AudioRewind":      vaxis.KeyMediaRewind,
	"XF86AudioNext":        vaxis.KeyMediaNext,
	"XF86AudioPrev":        vaxis.KeyMediaPrev,
	"XF86AudioRecord":      vaxis.KeyMediaRecord,
	"XF86AudioLowerVolume": vaxis.KeyMediaVolDown,
	"XF86AudioRaiseVolume": vaxis.KeyMediaVolUp,
	"XF86AudioMute":        vaxis.KeyMediaMute,
	"KP_0":                 vaxis.KeyKeyPad0,
	"KP_1":                 vaxis.KeyKeyPad1,
	"KP_2":                 vaxis.KeyKeyPad2,
	"KP_3":                 vaxis.KeyKeyPad3,
	"KP_4":                 vaxis.KeyKeyPad4,
	"KP_5":                 vaxis.KeyKeyPad5,
	"KP_6":                 vaxis.KeyKeyPad6,
	"KP_7":                 vaxis.KeyKeyPad7,
	"KP_8":                 vaxis.KeyKeyPad8,
	"KP_9":                 vaxis.KeyKeyPad9,
	"KP_Decimal":           vaxis.KeyKeyPadDecimal,
	"KP_Divide":            vaxis.KeyKeyPadDivide,
	"KP_Multiply":          vaxis.KeyKeyPadMultiply,
	"KP_Subtract":          vaxis.KeyKeyPadSubtract,
	"KP_Add":               vaxis.KeyKeyPadAdd,
	"KP_Enter":             vaxis.KeyKeyPadEnter,
	"KP_Equal":             vaxis.KeyKeyPadEqual,
	"KP_Separator":         vaxis.KeyKeyPadSeparator,
	"KP_Left":              vaxis.KeyKeyPadLeft,
	"KP_Right":             vaxis.KeyKeyPadRight,
	"KP_Up":                vaxis.KeyKeyPadUp,
	"KP_Down":              vaxis.KeyKeyPadDown,
	"KP_Page_Up":           vaxis.KeyKeyPadPageUp,
	"KP_Page_Down":         vaxis.KeyKeyPadPageDown,
	"KP_Home":              vaxis.KeyKeyPadHome,
	"KP_End":               vaxis.KeyKeyPadEnd,
	"KP_Insert":            vaxis.KeyKeyPadInsert,
	"KP_Delete":            vaxis.KeyKeyPadDelete,
	"KP_Begin":             vaxis.KeyKeyPadBegin,
	"Enter":                vaxis.KeyEnter,
	"Return":               vaxis.KeyReturn,
	"Tab":                  vaxis.KeyTab,
	"Escape":               vaxis.KeyEsc,

	// From xkb-commons, ASCII-only
	"space":        0x0020,
	"exclam":       0x0021,
	"quotedbl":     0x0022,
	"numbersign":   0x0023,
	"dollar":       0x0024,
	"percent":      0x0025,
	"ampersand":    0x0026,
	"apostrophe":   0x0027,
	"quoteright":   0x0027,
	"parenleft":    0x0028,
	"parenright":   0x0029,
	"asterisk":     0x002a,
	"plus":         0x002b,
	"comma":        0x002c,
	"minus":        0x002d,
	"period":       0x002e,
	"slash":        0x002f,
	"colon":        0x003a,
	"semicolon":    0x003b,
	"less":         0x003c,
	"equal":        0x003d,
	"greater":      0x003e,
	"question":     0x003f,
	"at":           0x0040,
	"bracketleft":  0x005b,
	"backslash":    0x005c,
	"bracketright": 0x005d,
	"asciicircum":  0x005e,
	"underscore":   0x005f,
	"grave":        0x0060,
	"quoteleft":    0x0060,
	"braceleft":    0x007b,
	"bar":          0x007c,
	"braceright":   0x007d,
	"asciitilde":   0x007e,
}
